import logging
from os import getpid
from platform import node

from minique.models.job import Job
from minique.models.queue import Queue
from minique.work.job_runner import JobRunner


class Worker:
    queue_timeout = 1

    def __init__(self, redis, queues):
        self.id = self.compute_id()
        self.redis = redis
        self.queues = queues
        self.log = logging.getLogger('{}.{}'.format(__name__, self.id.replace('.', '_')))
        assert all(isinstance(q, Queue) for q in queues)

    @classmethod
    def for_queue_names(cls, redis, queue_names, **kwargs):
        if isinstance(queue_names, str):
            queue_names = [queue_names]
        kwargs.setdefault('queues', [Queue(redis, name) for name in queue_names])
        return cls(redis, **kwargs)

    def compute_id(self):
        # Override me in a subclass if you like!
        return 'w-{node}-{pid}'.format(node=node().split('.')[0], pid=getpid())

    def get_next_job(self):
        rv = self.redis.blpop([q.redis_key for q in self.queues], self.queue_timeout)
        if rv:  # The rv is a 2-tuple (queue name, value)
            job_id = rv[1].decode()
            return Job(self.redis, job_id)

    def tick(self):
        job = self.get_next_job()
        if not job:
            return None
        runner = JobRunner(worker=self, job=job)
        runner.run()
        return job



import random
from contextlib import contextmanager
from importlib import import_module
from threading import local

_current_jobs = local()


def import_by_string(callable):
    module, _, func = callable.rpartition('.')
    module = import_module(module)
    func = getattr(module, func)
    return func


@contextmanager
def _set_current_job(job):
    assert not get_current_job()
    setattr(_current_jobs, 'current_job', job)
    try:
        yield
    finally:
        setattr(_current_jobs, 'current_job', None)


def get_current_job():
    return getattr(_current_jobs, 'current_job', None)


class cached_property(object):
    """A property that is only computed once per instance and then replaces
       itself with an ordinary attribute. Deleting the attribute resets the
       property.

       Source: https://github.com/bottlepy/bottle/blob/0.11.5/bottle.py#L175
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            # We're being accessed from the class itself, not from an object
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


consonants = 'cdgkklmmnnprst'
vowels = 'aeiou'


def get_random_pronounceable_string(length=12):
    s = []
    while length > 0 and len(s) < length:
        s.append(random.choice(consonants))
        s.append(random.choice(vowels) * random.choice((1, 1, 2)))
    return ''.join(s)[:length]



from minique.consts import QUEUE_KEY_PREFIX
from minique.utils import cached_property


class Queue:
    def __init__(self, redis, name):
        self.redis = redis
        self.name = name

    @cached_property
    def redis_key(self):
        return '%s%s' % (QUEUE_KEY_PREFIX, self.name)

    @property
    def length(self):
        return self.redis.llen(self.redis_key)



import json
import logging
import sys
import time
import traceback

from redis import StrictRedis

from minique.enums import JobStatus
from minique.excs import AlreadyAcquired, AlreadyResulted, NoSuchJob
from minique.models.job import Job
from minique.utils import _set_current_job, import_by_string


class JobRunner:
    def __init__(self, worker, job):
        assert isinstance(job, Job)
        self.worker = worker
        self.job = job
        self.redis = job.redis
        assert isinstance(self.redis, StrictRedis)
        self.log = logging.getLogger('{}.{}'.format(__name__, self.job.id.replace('.', '_')))

        if not job.exists:
            raise NoSuchJob('no such job: {id}'.format(id=self.job.id))

    def acquire(self):
        new_acquisition_info = json.dumps({'worker': self.worker.id, 'time': time.time()})
        if not self.redis.hsetnx(self.job.redis_key, 'acquired', new_acquisition_info):
            raise AlreadyAcquired('job {id} already acquired: {info}'.format(
                id=self.job.id,
                info=self.job.acquisition_info,
            ))
        self.redis.hset(self.job.redis_key, 'status', JobStatus.ACQUIRED.value)

    def execute(self):
        func = import_by_string(self.job.callable_name)
        kwargs = self.job.kwargs
        self.log.info('calling %s(%r)', func, kwargs)
        with _set_current_job(job=self.job):
            return func(**kwargs)

    def complete(self, success, value, duration):
        assert isinstance(success, bool)
        update_payload = {
            'status': (JobStatus.SUCCESS if success else JobStatus.FAILED).value,
            'duration': float(duration),
        }
        if not self.redis.setnx(self.job.result_redis_key, value):  # pragma: no cover
            raise AlreadyResulted('job {id} already has result'.format(id=self.job.id))
        self.redis.hmset(self.job.redis_key, update_payload)
        # Update expiries to the result TTL for both the job and the result
        self.redis.expire(self.job.result_redis_key, self.job.result_ttl)
        self.redis.expire(self.job.redis_key, self.job.result_ttl)
        if success:
            self.log.info('finished in %f seconds', duration)
        else:
            self.log.warning('errored in %f seconds', duration)

    def run(self):
        self.acquire()
        start_time = time.time()
        try:
            value = self.execute()
            value = json.dumps(value)
            success = True
        except Exception:
            success = False
            exc_type, exc_value, exc_tb = sys.exc_info()
            value = json.dumps({
                'exception_type': exc_type.__qualname__,
                'exception_value': str(exc_value),
                'traceback': traceback.format_exc(),
            })
        end_time = time.time()
        self.complete(success=success, value=value, duration=(end_time - start_time))



import json
import time

from minique.enums import JobStatus
from minique.excs import DuplicateJob
from minique.models.job import Job
from minique.models.queue import Queue
from minique.utils import get_random_pronounceable_string


def enqueue(redis, queue_name, callable, kwargs=None, job_id=None, job_ttl=0, result_ttl=86400 * 7):
    if not isinstance(callable, str):
        callable = '{module}.{qualname}'.format(
            module=callable.__module__,
            qualname=callable.__qualname__,
        )

    if not job_id:
        job_id = get_random_pronounceable_string()
    job = Job(redis, job_id)
    if job.exists:
        raise DuplicateJob('duplicate job: {id}'.format(id=job_id))
    payload = {
        'queue': queue_name,
        'callable': str(callable),
        'kwargs': json.dumps(kwargs or {}),
        'status': JobStatus.NONE.value,
        'ctime': time.time(),
        'job_ttl': int(job_ttl),
        'result_ttl': int(result_ttl),
    }
    queue = Queue(redis, name=queue_name)
    with redis.pipeline() as p:
        p.hmset(job.redis_key, payload)
        if payload['job_ttl'] > 0:
            p.expire(job.redis_key, payload['job_ttl'])
        p.rpush(queue.redis_key, job.id)
        p.execute()
        assert job.exists
    return job



import argparse
import logging

from redis import StrictRedis

from minique.work.worker import Worker

parser = argparse.ArgumentParser()
parser.add_argument('-u', '--redis-url', required=True)
parser.add_argument('-q', '--queues', nargs='+', required=True)


def main():
    args = parser.parse_args()
    logging.basicConfig(datefmt='%Y-%m-%d %H:%M:%S', level=logging.INFO)
    redis = StrictRedis.from_url(args.redis_url)
    worker = Worker.for_queue_names(redis=redis, queue_names=args.queues)
    worker.log.info('Worker initialized')
    while True:
        try:
            worker.tick()
        except KeyboardInterrupt:
            break
        except:
            worker.log.error('Unexpected worker tick error', exc_info=True)



import setuptools

dev_dependencies = [
    'flake8',
    'isort',
    'pydocstyle',
    'pytest',
    'pytest-cov',
]

if __name__ == '__main__':
    setuptools.setup(
        name='minique',
        description='Minimal Redis job runner',
        version='0.0.0',
        url='https://github.com/valohai/minique',
        author='Valohai',
        maintainer='Aarni Koskela',
        maintainer_email='akx@iki.fi',
        license='MIT',
        install_requires=['redis~=2.10.0'],
        tests_require=dev_dependencies,
        extras_require={'dev': dev_dependencies},
        packages=setuptools.find_packages('.', exclude=('minique_tests',)),
        include_package_data=True,
        entry_points={
            'console_scripts': [
                'minique = minique.cli:main',
            ],
        },
    )



import json

from redis import StrictRedis

from minique.consts import JOB_KEY_PREFIX, RESULT_KEY_PREFIX
from minique.enums import JobStatus


class Job:
    def __init__(self, redis: StrictRedis, id):
        self.redis = redis
        assert isinstance(id, str)
        self.id = id

    @property
    def redis_key(self):
        return '%s%s' % (JOB_KEY_PREFIX, self.id)

    @property
    def result_redis_key(self):
        return '%s%s' % (RESULT_KEY_PREFIX, self.id)

    @property
    def acquisition_info(self):
        return json.loads(self.redis.hget(self.redis_key, 'acquired').decode())

    @property
    def has_finished(self):
        return self.redis.exists(self.result_redis_key)

    @property
    def result(self):
        # TODO: exception handling
        return json.loads(self.redis.get(self.result_redis_key).decode())

    @property
    def status(self):
        return JobStatus(self.redis.hget(self.redis_key, 'status').decode())

    @property
    def exists(self):
        return self.redis.exists(self.redis_key)

    @property
    def result_ttl(self):
        return int(self.redis.hget(self.redis_key, 'result_ttl'))

    @property
    def duration(self):
        return float(self.redis.hget(self.redis_key, 'duration'))

    @property
    def kwargs(self):
        # TODO: exception handling
        return json.loads(self.redis.hget(self.redis_key, 'kwargs').decode())

    @property
    def callable_name(self):
        return self.redis.hget(self.redis_key, 'callable').decode()

    def __str__(self):
        return '<job %s>' % self.id

    def __eq__(self, other):
        return isinstance(other, Job) and (self.id == other.id)
